#ifndef _CONTROLE_H_
#define _CONTROLE_H_

#include "futdados.h"

class PID {
 private:
  double K, Ti, Td, N;
  double e_ant, I_ant, I_ant2, D_ant;
 public:
  // A relao com as constantes tradicionais do PID  a seguinte:
  // Kp=k, Ki=k/ti, Kd=k*td
  // Para eliminar a parte integral (Ki=0), ti=infinito
  // Para eliminar a parte derivativa (Kd=0), td=0.0
  // O nmero n  o limitador de altas freqencias da parte derivativa.
  // Normalmente, 3<=n<=20. Para eliminar a filtragem, n=infinito
  inline PID(double k=1.0, double ti=1E+6, double td=0.0, double n=1E+6) {
    fixa_constantes(k,ti,td,n);
  }
  void fixa_constantes(double k, double ti, double td, double n=1E+6);
  void anti_windup();
  void reset();
  double controle(double erro, double h);
};

class Controle:public virtual Futdados
{
 private:
  bool controle_orientacao;
  bool chegou[3];
  PID lin[3],ang[3];
 public:
  Controle(TIME time);
  bool controle();
  bool suspender_controle();
};

#endif
