#include <iostream>
#include <cmath>
#include "parametros.h"
#include "funcoes.h"
#include "obstaculos.h"

bool fusao_obst(double &x1, double &y1, double &r1,
		double x2, double y2, double r2);
double desvio_obst(double x_rob, double y_rob, double x_ref, double y_ref,
		   double x_obs, double y_obs, double raio);

Obstaculos::Obstaculos(TIME time): Futdados(time)
{
}

bool Obstaculos::obstaculos()
{
  bool ignora_meu[3], ignora_adv[3];
  double dist,xc,yc,raio,gama,delta_gama;
  int i,j,k;

  // Calcula o efeito dos obstculos sobre o i-simo rob do meu time
  for (i=0; i<3; i++) {
    delta_gama = 0.0;
    for (j=0; j<3; j++) {
      // A princpio, todos os robs do meu time so obstculos
      ignora_meu[j] = false;
      // Todos os robs do outro time so obstculos, exceto o
      // adversrio que est com a bola
      dist = hypot(pos.bola.x-pos.adv[j].x, pos.bola.y-pos.adv[j].y);
      ignora_adv[j] = (dist<=RAIO_ROBO+RAIO_BOLA);
    }
    // O rob no  obstculo para si mesmo
    ignora_meu[i] = true;
    // Calcula o desvio dos outros robs do meu time
    for (j=0; j<3; j++) if (!ignora_meu[j]) {
      // Inicialmente, o j-simo rob  representado por um obstculo
      // posicionado na mesma posio e com o mesmo raio que o rob
      xc = pos.meu[j].x;
      yc = pos.meu[j].y;
      raio = RAIO_ROBO;
      // Testa se o j-simo rob do meu time deve ser fundido com o
      // k-simo rob do mesmo time
      for (k=j+1; k<3; k++) if (!ignora_meu[k]) {
	ignora_meu[k] = fusao_obst(xc,yc,raio,
				   pos.meu[k].x,pos.meu[k].y,RAIO_ROBO);
      }
      // Testa se o j-simo rob do meu time deve ser fundido com o
      // k-simo rob do outro time
      for (k=0; k<3; k++) if (!ignora_adv[k]) {
	ignora_adv[k] = fusao_obst(xc,yc,raio,
				   pos.adv[k].x,pos.adv[k].y,RAIO_ROBO);
      }
      // Desvia do j-simo rob do meu time
      gama = desvio_obst(pos.meu[i].x,pos.meu[i].y,
			 ref.meu[i].x,ref.meu[i].y,
			 xc,yc,raio);
      if (fabs(gama)>fabs(delta_gama)) delta_gama = gama;
    }
    // Calcula o desvio dos robs do outro time
    for (j=0; j<3; j++) if (!ignora_adv[j]) {
      // Inicialmente, o j-simo rob  representado por um obstculo
      // posicionado na mesma posio e com o mesmo raio que o rob
      xc = pos.adv[j].x;
      yc = pos.adv[j].y;
      raio = RAIO_ROBO;
      // Testa se o j-simo rob do outro time deve ser fundido com o
      // k-simo rob do mesmo time
      for (k=j+1; k<3; k++) if (!ignora_adv[k]) {
	ignora_adv[k] = fusao_obst(xc,yc,raio,
				   pos.adv[k].x,pos.adv[k].y,RAIO_ROBO);
      }
      // Desvia do j-simo rob do outro time
      gama = desvio_obst(pos.meu[i].x,pos.meu[i].y,
			 ref.meu[i].x,ref.meu[i].y,
			 xc,yc,raio);
      if (fabs(gama)>fabs(delta_gama)) delta_gama = gama;
    }
    // Desvia a referncia do i-simo rob
    gama = arc_tang(ref.meu[i].y-pos.meu[i].y,ref.meu[i].x-pos.meu[i].x);
    dist = hypot(ref.meu[i].y-pos.meu[i].y,ref.meu[i].x-pos.meu[i].x);
    gama += delta_gama;
    ref.meu[i].x = pos.meu[i].x + dist*cos(gama);
    ref.meu[i].y = pos.meu[i].y + dist*sin(gama);
    ref.meu[i].theta += delta_gama;
  }
  return false;
}

// Caso a distncia entre dois obstculos no permita a passagem de um
// rob entre eles, os dois obstculos so fundidos no primeiro.
// Retorna true caso a fuso tenha ocorrido

bool fusao_obst(double &x1, double &y1, double &r1,
		double x2, double y2, double r2)
{
  double dist = hypot(y2-y1, x2-x1);
  if (dist-r1-r2 > 2*RAIO_ROBO ) {
    return false;
  }
  double theta = arc_tang(y2-y1, x2-x1);
  double novo_raio = (dist+r1+r2)/2.0;
  x1 += (novo_raio-r1)*cos(theta);
  y1 += (novo_raio-r1)*sin(theta);
  r1 = novo_raio;
  return true;
}

// Calcula de quanto deve ser girada a referncia do rob, dada a
// posio do rob, a referncia atual e a posio e o raio do obstculo

double desvio_obst(double x_rob, double y_rob, double x_ref, double y_ref,
		   double x_obs, double y_obs, double raio)
{
  raio = fabs(raio)+RAIO_ROBO;  // Obstculo engrandecido
  double delta_x=x_obs-x_rob, delta_y=y_obs-y_rob;
  double dist = hypot(delta_x,delta_y);
  if (dist-raio >= ZONA_INFL_MAX) {
    // O rob est fora da zona de influncia do obstculo
    return 0.0;
  }
  double coef;
  if (dist-raio <= ZONA_INFL_MIN) {
    // O rob est dentro da zona de influncia total do obstculo
    coef = 1.0;
  }
  else {
    // O rob est dentro da zona de influncia parcial do obstculo
    coef = 1.0-(dist-raio-ZONA_INFL_MIN)/(ZONA_INFL_MAX-ZONA_INFL_MIN);
  }

  double ang_faixa,t1,t2;
  // Calcula a faixa angular [t1,t2] que est "eclipsada" pelo obstculo
  if (dist >= raio) {
    ang_faixa = asin(raio/dist);
  }
  else {
    dist = 2*raio-dist;
    ang_faixa = M_PI-asin(raio/dist);
  }
  t2 = ang_equiv(arc_tang(delta_y,delta_x) + ang_faixa);
  t1 = ang_equiv(arc_tang(delta_y,delta_x) - ang_faixa);
  // Calcula o ngulo do rob para a referncia. Se estiver na faixa
  // "eclipsada", retorna o ngulo de giro.
  double theta=arc_tang(y_ref-y_rob,x_ref-x_rob);
  if (t2<t1) {
    if (theta>=0) t2 += 2*M_PI;
    else t1 -= 2*M_PI;
  }
  if (theta>=t2 || theta<=t1) {
    return 0.0;
  }
  double gama;
  if (t2-theta < theta-t1) {
    gama = t2-theta;
  }
  else {
    gama = t1-theta;
  }
  return coef*gama;
}

