#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <sys/sem.h>
#include <sys/shm.h>
#include <math.h>

#include "ssocket.h"
#include "dados.h"
#include "comunicacao.h"
#include "sistema.h"
#include "funcoes.h"
#include "modelo.h"

using namespace std;

/* ***************************************************************
   Funcoes
   *************************************************************** */

static void* controle_soquetes(void*);  // main da thread dos sockets
static void* robos(void*);              // main da thread de simulacao
static void atualiza_controle(CONTROLE&, TIME, Modelo &);

/* ****************************************************************
   Variaveis globais (das duas threads)
   **************************************************************** */

static Modelo est;
static SITUACAO *psit = NULL;
static int sem_compart = -1;
static int mem_compart = -1;
static EST_JOGO estado_simulacao = ATIVO;

// sockets de comunicacao com os clientes
udpSocket socket_azul,socket_amrl;
// filas de controle (para teste da existencia de dados a ler)
fsocket filaServidores,filaClientes;

/* ****************************************************************
   PROGRAMA PRINCIPAL
   ***************************************************************** */

static void erro(SOCKET_STATUS err, const char *msg){}

int main(void)
{
  // Para que a biblioteca de sockets nao fa'ca nada em caso de "erros"
  // (por exemplo, de timeouts). Ela deve apenas retornar o codigo de
  // erro, para que eu trate aqui no programa principal.
  socket_error = erro;

  // Threads
  pthread_t thrd_controle;
  pthread_t thrd_simular;
  SITUACAO minhaSit;

  cout << "****************************************************\n";
  cout << "* Programa simulador dos robo do futebol de robo *\n";
  cout << "****************************************************\n";

  /* ************************************************
     Criacao dos semaforos, memoria compart e threads
     ************************************************ */

  // Criacao do semaforo de exclusao mtua do compartilhamento dos dados
  if ((sem_compart=semget(KEY_SEM_COMPART,1,IPC_EXCL|IPC_CREAT|0600)) == -1) {
    cerr << "Erro na criacao do semaforo de compartilhamento\n";
    goto FIM;
  }
  if (semctl(sem_compart,0,SETVAL,1) == -1) {
    cerr << "Erro na inicializacao do semaforo de compartilhamento\n";
    goto FIM;
  }

  // Criacao da memoria compartilhada para o compartilhamento de dados
  if ((mem_compart=shmget(KEY_MEM_COMPART,sizeof(SITUACAO),
			  IPC_EXCL|IPC_CREAT|0600))==-1) {
    cerr << "Erro na criacao da memoria compartilhada\n";
    goto FIM;
  }
  if ((psit=(SITUACAO *)shmat(mem_compart,NULL,0)) == (void*)-1) {
    cerr << "Erro no acoplamento da memoria compartilhada\n";
    goto FIM;
  }

  // Criacao da thread de controle dos soquetes
  // verifica conexoes
  if (pthread_create(&thrd_controle, NULL, controle_soquetes, NULL) != 0) {
    cerr << "Erro na criacao da thread de controle das conexoes\n";
    goto FIM;
  }
  if (pthread_detach(thrd_controle) != 0) {
    cerr << "Erro na dissociacao da thread de conexoes\n";
    goto FIM;
  }

  // Criacao da thread de simulacao (simulador <-> controle),
  // que engloba as funcoes de envio (camera), recepcao (porta paralela)
  // e simular_modelo (modelo dos robos e modelagem de colisoes)
  if (pthread_create(&thrd_simular, NULL, robos, NULL) != 0) {
    cerr << "Erro na criacao da thread de simulacao\n";
    goto FIM;
  }
  if (pthread_detach(thrd_simular) != 0) {
    cerr << "Erro na dissociacao da thread de simulacao\n";
    goto FIM;
  }
  
  /* *********************************
     Processo pai (programa principal)
     ********************************* */

  sleep(1);       // Para dar tempo as threads de se lan'carem
  set_keypress(); // Para nao precisar digitar <ENTER> apos a tecla

  do {
    char tecla;

    cout << "\nOpcoes disponiveis:\n";
    cout << "  I - Voltar ao estado (i)nicial\n";
    cout << "  P - Im(p)rimir estado\n";
    cout << "  B - (B)ola-Livre\n";
    cout << "  G - (G)ol\n";
    cout << "\n";
    cout << "  A - Penalti para time Azul\n";
    cout << "  S - Tiro-Livre para time Azul\n";
    cout << "  D - Tiro-de-Meta para time Azul\n";
    cout << "\n";
    cout << "  Z - Penalti para time Amarelo\n";
    cout << "  X - Tiro-Livre para time Amarelo\n";
    cout << "  C - Tiro-de-Meta para time Amarelo\n";
    cout << "\n";
//    cout << "S - (S)uspender simulacao\n";
//    cout << "C - (C)ontinuar simulacao\n";
    cout << "  T - (T)erminar\n";
    cout << "SUA OPCAO: "; 
    do {
      tecla = getchar();
    } while (tecla==EOF && estado_simulacao!=TERMINAR);

    if (estado_simulacao==TERMINAR) {
      tecla = 'T';
      cout << tecla;
    }
    else {
      tecla = toupper(tecla);
    }
    cout << '\n';
    switch(tecla) {
    case 'I': 
      est.posicao_inicial();
      break;
    case 'P':
      // Imprime posicoes dos robos e da bola
      // A informacao eh lida na memoria compartilhada
      if (sem_compart!=-1 && entrar_regiao_critica(sem_compart)) {
	cerr << "Erro obtendo acesso a memoria compartilhada\n";
	sem_compart = -1;
      }
      if (sem_compart!=-1 && psit!=NULL) {
	// Ler na memoria compartilhada
	memcpy(&minhaSit,psit,sizeof(SITUACAO));
      }
      if (sem_compart!=-1 && sair_regiao_critica(sem_compart)) {
	cerr << "Erro liberando acesso a memoria compartilhada\n";
	sem_compart = -1;
      }
      // Escrevendo a informacao na tela
      cout << "+------------------------------------"
	   << "-----------------------------------+\n";
      printf( "| %6lu ", minhaSit.id);
//      printf( "|           AZUL       %3u ", minhaSit.gols.azul);
//      printf( "x %-3u      AMARELO         ", minhaSit.gols.amrl);
      printf( "|           AZUL           ");
      printf( "|        AMARELO           ");
      printf( "|        |\n");
//      printf( "|        |\n");
      // Nunca se imprime quando o estado eh TERMINAR
      printf( "| %6.6s ", estado_simulacao==ATIVO ? "ATIVO" : "SUSPENSO");
      cout << "| 0 CIAN | 1 ROSA | 2 VERD "
	   << "| 0 CIAN | 1 ROSA | 2 VERD "
	   << "|   BOLA |\n";
      cout << "+------------------------------------"
	   << "-----------------------------------+\n";
      printf( "| Pos x  ");
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.azul[0].x,
	      minhaSit.pos.azul[1].x, minhaSit.pos.azul[2].x);
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.amrl[0].x,
	      minhaSit.pos.amrl[1].x, minhaSit.pos.amrl[2].x);
      printf( "| %+6.3f |\n", minhaSit.pos.bola.x);
      printf( "| Pos y  ");
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.azul[0].y,
	      minhaSit.pos.azul[1].y, minhaSit.pos.azul[2].y);
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.amrl[0].y,
	      minhaSit.pos.amrl[1].y, minhaSit.pos.amrl[2].y);
      printf( "| %+6.3f |\n", minhaSit.pos.bola.y);
      printf( "| Theta  ");
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.azul[0].theta,
	      minhaSit.pos.azul[1].theta, minhaSit.pos.azul[2].theta);
      printf( "| %+6.3f | %+6.3f | %+6.3f ", minhaSit.pos.amrl[0].theta,
	      minhaSit.pos.amrl[1].theta, minhaSit.pos.amrl[2].theta);
      printf( "| ------ |\n");
      cout << "+------------------------------------"
	   << "-----------------------------------+\n";
      break;
    case 'T': 
    //case 'S':
    //case 'C':
      /*if (tecla=='C') {
	if (estado_simulacao==SUSPENSO) est.set_tempo(relogio());
	estado_simulacao=ATIVO;
      }
      else {*/
	//estado_simulacao=(tecla=='S' ? SUSPENSO : TERMINAR);
	estado_simulacao = TERMINAR;
      //}
      cout << "+----------------------------------+\n";
      cout << "| ESTADO: ";
      switch (estado_simulacao) {
      case ATIVO:
	cout << "Simulacao ativa          |\n";
	break;
      case SUSPENSO: //Nao deve ocorrer pois o simulador nao fica mais suspenso..
	cout << "Simulacao suspensa       |\n";
	break;
      case TERMINAR:
	cout << "Simulacao terminada      |\n";
	break;
      }
      cout << "+----------------------------------+\n";
      break;
    case 'B':
      est.bola_livre();
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'G':
      est.bola_ao_centro();
      break;
    case 'A':
      est.penalti(TIME_AZUL);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'S':
      est.penalti(TIME_AZUL);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'D':
      est.tiro_de_meta(TIME_AZUL);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'Z':
      est.penalti(TIME_AMARELO);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'X':
      est.penalti(TIME_AMARELO);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    case 'C':
      est.tiro_de_meta(TIME_AMARELO);
      //cout << "\nOpcao nao implementada!\n\n";
      break;
    }
  } while (estado_simulacao!=TERMINAR);

  //se estado_simulacao == terminado ou acontecer qualquer erro na criacao 
  //de semaforos, memoria compartilhada ou threads, fim do simulador.
 FIM:
  estado_simulacao=TERMINAR;
  reset_keypress();

  // Desacoplamento e destruicao da memoria compartilhada
  if (psit != NULL) {
    if (shmdt(psit) == -1) {
      cerr << "Erro no desacoplamento da memoria compartilhada\n";
    }
    psit = NULL;
  }
  if (mem_compart != -1) {
    if (shmctl(mem_compart,IPC_RMID,NULL) == -1) {
      cerr << "Erro na destruicao da memoria compartilhada\n";
    }
    mem_compart = -1;
  }

  // Destruicao do semaforo de compartilhamento
  if (sem_compart != -1) {
    if (semctl(sem_compart,0,IPC_RMID,0) == -1) {
      cerr << "Erro na destruicao do semaforo de compartilhamento\n";
    }
    sem_compart = -1;
  }
}

static void* controle_soquetes(void *pt_c)
{
  int servidoresAtivos;
  SOCKET_STATUS result;

  /* ************************************************
     Espera pela conexao dos sockets
     ************************************************ */

  while (estado_simulacao != TERMINAR) {
    // Inicializacao dos sockets de controle
    if (socket_azul.closed() &&
	socket_azul.listen( PORTA_AZUL ) != SOCKET_OK) {
      cerr << "Falha na abertura do servidor azul\n";
      estado_simulacao=TERMINAR;
    }
    if (socket_amrl.closed() &&
	socket_amrl.listen( PORTA_AMRL ) != SOCKET_OK) {
      cerr << "Falha na abertura do servidor amarelo\n";
      estado_simulacao=TERMINAR;
    }

    filaServidores.clean();
    servidoresAtivos = 0;
    if (socket_azul.accepting()) {
      filaServidores.include(socket_azul);
      servidoresAtivos++;
    }
    if (socket_amrl.accepting()) {
      filaServidores.include(socket_amrl);
      servidoresAtivos++;
    }
    if (servidoresAtivos > 0) {
      result = filaServidores.wait_connect(1000);  // 1000 ms
      if( result == SOCKET_OK ) {
	if( socket_azul.accepting() &&
	    filaServidores.had_activity(socket_azul) ) {
	  if( socket_azul.accept() == SOCKET_OK ) {
	    cout << "Conexao aceita com time azul\n"; 
	  }  // fim do if( socket_azul.accept() == SOCKET_OK )
	  else {
	    cerr << "Conexao nao aceita pelo servidor azul\n";
	    estado_simulacao=TERMINAR;
	  }
	} // fim if( filaServidores.had_activity(servidor_azul) )
	if( socket_amrl.accepting() &&
	    filaServidores.had_activity(socket_amrl) ) {
	  if( socket_amrl.accept() == SOCKET_OK ) {
	    cout << "Conexao aceita com time amarelo\n"; 
	  }  // fim if( socket_amrl.accept() == SOCKET_OK )
	  else {
	    cerr << "Conexao nao aceita pelo servidor amarelo\n";
	    estado_simulacao=TERMINAR;
	  }
	} // fim if( filaServidores.had_activity(servidor_amrl) )
      } 
      else if (result != SOCKET_TIMEOUT) {  // else if ( result == SOCKET_OK )
	cerr << "Erro na fila de sockets servidores\n";
	estado_simulacao=TERMINAR;
      } // fim if( result == SOCKET_OK )
    }
    else { // else if( servidoresAtivos > 0 )
      sleep(1);
    } // fim if( servidoresAtivos > 0 )
  } // fim_while

  socket_azul.close();
  socket_amrl.close();

  return NULL;
}

/* Simulacao (envio+modelo+recepcao). Esta eh a funcao "main" da thread
   mais importante, que faz efetivamente a simulacao. */

static void* robos(void *pt_c)
{
  SITUACAO minhaSit;
  SINAL_RADIO ctrl_azul, ctrl_amrl, *meuCtrl;
  udpSocket *meuSocket;
  SOCKET_STATUS result;
  //double t,dt,dsusp,damostr=0.0;
  double t,dt,damostr=0.0;
  int clientesAtivos;

  est.posicao_inicial();
  minhaSit.id = 0;      // Vai gerar a primeira imagem
//  minhaSit.gols.azul = minhaSit.gols.amrl = 0;
  //ctrl_azul.id = ctrl_amrl.id = 0;   // Nao recebeu nenhum controle
  damostr = 0.0;

  /* ***************************************
     laco (enquanto simulador estiver ativo)
     *************************************** */
  while (estado_simulacao!=TERMINAR) {
    // Le a hora atual
    t = relogio();

    // Fila de transmissores de sinais de controle que estao conectados
    filaClientes.clean();
    clientesAtivos = 0;
    if( socket_azul.connected() ) {
      filaClientes.include(socket_azul);
      clientesAtivos++;
    }
    if( socket_amrl.connected() ) {
      filaClientes.include(socket_amrl);
      clientesAtivos++;
    }
    // Leitura de eventuais sinais de controle
    if (clientesAtivos > 0) {
      result = filaClientes.wait_read(0);  // timeout=0
      if ( result == SOCKET_OK ) {
	for( int ii=0; ii<2; ii++ ) {
	  meuCtrl = (ii==0 ? &ctrl_amrl : &ctrl_azul);
	  meuSocket = (ii==0 ? &socket_amrl : &socket_azul);
	  if( meuSocket->connected() &&
	      filaClientes.had_activity(*meuSocket) ) {
	    do {
	      result = meuSocket->read((void *)meuCtrl, sizeof(SINAL_RADIO));
	      if( result == SOCKET_OK ) {
		fsocket fila;
		fila.include(*meuSocket);
		result = fila.wait_read(0);
		if( result != SOCKET_TIMEOUT && result != SOCKET_OK ) {
		  cerr << "Erro na fila interna de socket" << endl;
		  estado_simulacao=TERMINAR;
		  return (void*)-1;
		}
	      }
	      else {
		cout << "Conexao fechada com transmissor do time "
		   << ((ii==0)?"amarelo":"azul")
		   << endl;
		meuSocket->close();
	      }
	   /*   if (result==SOCKET_OK) {
		cerr << "Sinal de controle desprezado: "
		     << meuCtrl->id << endl;
	      }*/
	    } while( result == SOCKET_OK );
	    atualiza_controle(meuCtrl->c, 
			      (ii==0) ? TIME_AMARELO : TIME_AZUL, est);
/*	    if (meuCtrl->id+1 != minhaSit.id) {
	      cerr << "Controle atrasado para o time "
		   << ((ii==0)?"amarelo":"azul")
		   << ": ctrl=" << meuCtrl->id
		   << ": img=" << minhaSit.id << endl;
	    }*/
	  } // fim if( meuSocket->connected() && ...
	}// fim for
      } // fim if ( result == SOCKET_OK ) {
      else if( result != SOCKET_TIMEOUT ) {
	cerr << "Erro na fila de sockets" << endl;
	estado_simulacao=TERMINAR;
	return (void*)-1;
      }

    } // fim if (clientesAtivos != 0)

    // Intervalo desde o ltimo ciclo
    dt = t-est.le_tempo();
    if (estado_simulacao==ATIVO) {
      est.simular(dt);
    }
/*    else {  // SUSPESO
      est.avancar_sem_simular(dt);
      dsusp += dt;
    }*/
    damostr += dt;
    //CONFERIR COM ADELARDO!!!!!
/*    if (estado_simulacao==SUSPENSO) {
      dsusp += dt;
    }
    else {*/
//      dsusp = 0.0;
    //}

    // Envio de imagem a cada periodo de amostragem
    if ( damostr >= T_AMOSTR ) {
//      minhaSit.estado = estado_simulacao;
      est.le_posicao(minhaSit.pos);
      //normalizacao dos angulos
      for (int i=0; i<3; i++) {
	minhaSit.pos.azul[i].theta = ang_equiv(minhaSit.pos.azul[i].theta);
	minhaSit.pos.amrl[i].theta = ang_equiv(minhaSit.pos.amrl[i].theta);
      }
      // Escrever na memoria compartilhada (para visualizador)
      if (estado_simulacao!=TERMINAR && sem_compart!=-1 &&
	  entrar_regiao_critica(sem_compart)) {
	cerr << "Erro obtendo acesso a memoria compartilhada\n";
	sem_compart = -1;
      }
      if (estado_simulacao!=TERMINAR && sem_compart!=-1 && psit!=NULL) {
	// Escrever na memoria compartilhada
	memcpy(psit, &minhaSit, sizeof(SITUACAO));
      }
      if (estado_simulacao!=TERMINAR && sem_compart!=-1 &&
	  sair_regiao_critica(sem_compart)) {
	cerr << "Erro liberando acesso a memoria compartilhada\n";
	sem_compart = -1;
      }
      // Se houver algum cliente conectado, manda imagem nova
      if ( socket_amrl.connected() ) {
	if (socket_amrl.write((void*)&minhaSit,
			       sizeof(SITUACAO))!=SOCKET_OK) {
	  cout << "Conexao fechada com camera amarela\n";
	  socket_amrl.close();
	}
      }
      if ( socket_azul.connected() ) {
	if (socket_azul.write((void*)&minhaSit,
			       sizeof(SITUACAO))!=SOCKET_OK) {
	  cout << "Conexao fechada com camera azul\n";
	  socket_azul.close();
	}
      }
      damostr -= T_AMOSTR;
      minhaSit.id++;  // id do quadrado gerado
    } // fim if ( damostr >= T_AMOSTR )

    // Verifica se houve gols marcados
/*
    if (estado_simulacao == ATIVO &&
	fabs(minhaSit.pos.bola.x) > LARGURA_CAMPO/2.0) {
      if (minhaSit.pos.bola.x > 0.0) {
	minhaSit.gols.azul++;
      }
      else {
	minhaSit.gols.amrl++;
      }
      estado_simulacao = minhaSit.estado = SUSPENSO;
      est.posicao_inicial();
    }
*/
    // Verifica se ja deve sair do estado SUSPENSO
/*    if (estado_simulacao == SUSPENSO && dsusp > 10) {
      estado_simulacao = minhaSit.estado = ATIVO;
      dsusp = 0.0;
    }*/

    usleep(9999); // 10ms = tick do Linux
  } //fim while (estado_simulacao != TERMINAR)
  return(NULL);
}

// atualiza sinal de controle
static void atualiza_controle(CONTROLE &controle, TIME cor, Modelo &est)
{
  for (int id=0; id<3; id++) {
    // Esquerdo
    est.atualizar_controle(cor, id, MOTOR_ESQUERDO, controle.meu[id].left);
    // Direito
    est.atualizar_controle(cor, id, MOTOR_DIREITO, controle.meu[id].right);
  }
}
